/* 
**  mod_deny_deeplink.c -- Apache module which disallow DeepLink.
**
**  To play with this sample module, first compile it into a
**  DSO file and install it into Apache's libexec directory 
**  by running:
**
**    $ apxs -c -i mod_deny_deeplink.c
**
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "ap_config.h"
#include "http_log.h"
#include "ap_sha1.h"

#include "config.h"


enum {
    DISABLE   = 0,
    ENABLE    = 1,
    UNDEFINED = -1,
};

static const char VERSION_STRING[] = PACKAGE_NAME "/" PACKAGE_VERSION;
static const char DEFAULT_REDIRECT_URL[] = "http://www.sankei.co.jp/";
static const char DEFAULT_SECRET_KEY[]   = "mod_deny_deeplink.c secret key";
static const int  TIMEOUT                = 60;
static const char DEFAULT_COOKIE_NAME[]  = "mod_deny_deeplink";
static const int  HMAC_BLOCKSIZE         = 64;

typedef struct {
    int enable;
    const char *redirect_url;
    const char *secret;
    int starting_point;
    const char *cookie_name;
    int timeout;
} deny_deeplink_rec;

module MODULE_VAR_EXPORT deny_deeplink_module;


static void init_module(server_rec *s, pool *p)
{
    ap_add_version_component(VERSION_STRING);
}


static void *create_dir_config(pool *p, char *path)
{
    deny_deeplink_rec *conf;

    conf = ap_pcalloc(p, sizeof(deny_deeplink_rec));
    conf->enable         = UNDEFINED;
    conf->redirect_url   = NULL;
    conf->starting_point = UNDEFINED;
    conf->secret         = NULL;
    conf->cookie_name    = NULL;
    conf->timeout        = UNDEFINED;
    return (void *)conf;
}


static void *merge_dir_config(pool *p, void *basev, void *addv)
{
    deny_deeplink_rec *base = (deny_deeplink_rec *)basev;
    deny_deeplink_rec *add  = (deny_deeplink_rec *)addv;
    deny_deeplink_rec *new;

    new = ap_pcalloc(p, sizeof(deny_deeplink_rec));
    new->enable       = base->enable;
    new->redirect_url = ap_pstrdup(p, base->redirect_url);
    new->secret       = ap_pstrdup(p, base->secret);
    new->cookie_name  = ap_pstrdup(p, base->cookie_name);
    new->timeout      = base->timeout;

    if (add->starting_point != UNDEFINED)
        new->starting_point = add->starting_point;
    if (add->enable != UNDEFINED)
        new->enable = add->enable;
    if (add->redirect_url != NULL)
        new->redirect_url = ap_pstrdup(p, add->redirect_url);
    else
        new->redirect_url = ap_pstrdup(p, base->redirect_url);

    if (add->secret != NULL)
        new->secret = ap_pstrdup(p, add->secret);
    else
        new->secret = ap_pstrdup(p, base->secret);

    if (add->cookie_name != NULL)
        new->cookie_name = ap_pstrdup(p, add->cookie_name);
    else
        new->cookie_name = ap_pstrdup(p, base->cookie_name);

    if (add->timeout != UNDEFINED)
        new->timeout = add->timeout;

    return new;
}


static const char *cmd_timeout(cmd_parms *parms, void *mconfig, char *timeout)
{
    deny_deeplink_rec *conf;

    if (timeout == NULL)
        return NULL;
    conf = mconfig;
    conf->timeout = atoi((const char *)timeout);

    return NULL;
}


static const command_rec commands[] =
{
    { "DenyDeeplink", ap_set_flag_slot,
      (void *)XtOffsetOf(deny_deeplink_rec, enable),
      OR_ALL, TAKE1, "deny deeplink if On" },
    { "DenyDeeplinkStartingPoint", ap_set_flag_slot,
      (void *)XtOffsetOf(deny_deeplink_rec, starting_point),
      OR_ALL, TAKE1, "deny deeplink starting point if On" },
    { "DenyDeeplinkRedirectURL", ap_set_string_slot,
      (void *)XtOffsetOf(deny_deeplink_rec, redirect_url),
      OR_ALL, TAKE1, "deny deeplink redirect URL" },
    { "DenyDeeplinkSecretKey",  ap_set_string_slot,
      (void *)XtOffsetOf(deny_deeplink_rec, secret),
      OR_ALL, TAKE1, "deny deeplink secret key for HMAC" },
    { "DenyDeeplinkCookieName", ap_set_string_slot,
      (void *)XtOffsetOf(deny_deeplink_rec, cookie_name),
      OR_ALL, TAKE1, "deny deeplink HTTP Cookie name" },
    { "DenyDeeplinkTimeout", cmd_timeout,
      NULL,
      OR_ALL, TAKE1, "deny deeplink timeout" },
    { NULL }
};


static char *get_hmac_string(pool *p, char *msg, const unsigned char *key)
{
    AP_SHA1_CTX ctx;
    unsigned char digest[SHA_DIGESTSIZE];
    unsigned char k_ipad[65], k_opad[65];
    unsigned char kt[SHA_DIGESTSIZE];
    unsigned char *k = (unsigned char *)key;
    char hex_digest[SHA_DIGESTSIZE * 2], *hd;
    const char hex[] = "0123456789abcdef";
    int key_len, i;

    key_len = strlen(key);
    /* if key is longer than 64 bytes reset it to key=SHA1(key) */
    if (key_len > 64) {
        k = kt;
        key_len = SHA_DIGESTSIZE; 
    }

    /* HMAC_MD5 transform */
    /* start out by string key in pads */
    memset((void *)k_ipad, 0, sizeof(k_ipad));
    memset((void *)k_opad, 0, sizeof(k_opad));
    memmove(k_ipad, k, key_len);
    memmove(k_opad, k, key_len);

    /* XOR key with ipad and opad values */
    for (i = 0; i < 64; i++) {
        k_ipad[i] ^= 0x36;
        k_opad[i] ^= 0x5c;
    }
    /* perform inner SHA-1 */
    ap_SHA1Init(&ctx);
    ap_SHA1Update_binary(&ctx, k_ipad, 64);
    ap_SHA1Update_binary(&ctx, (const unsigned char *)msg, strlen(msg));
    ap_SHA1Final(digest, &ctx);
    /* perform outer SHA-1 */
    ap_SHA1Init(&ctx);
    ap_SHA1Update_binary(&ctx, k_opad, 64);
    ap_SHA1Update_binary(&ctx, digest, sizeof(digest));
    ap_SHA1Final(digest, &ctx);

    hd = hex_digest;
    for (i = 0; i < sizeof(digest); i++) {
        *hd++ = hex[digest[i] >> 4];
        *hd++ = hex[digest[i] & 0xF];
    }
    *hd = '\0';

    return ap_pstrdup(p, hex_digest);
}


static int set_cookie(request_rec *r)
{
    deny_deeplink_rec *conf;
    const char *name;
    char *value, *cookie, *timestr, *hmac;
    time_t timeout;

    conf = ap_get_module_config(r->per_dir_config, &deny_deeplink_module);
    timeout = time(NULL) + (conf->timeout != UNDEFINED ? conf->timeout : TIMEOUT);
    timestr = ap_psprintf(r->pool, "%u", (unsigned int)timeout);    
    hmac = get_hmac_string(r->pool, timestr,
                           conf->secret ? conf->secret : DEFAULT_SECRET_KEY);

    name = conf->cookie_name ? conf->cookie_name : DEFAULT_COOKIE_NAME;
    value = ap_pstrcat(r->pool, hmac, timestr, NULL);
    cookie = ap_pstrcat(r->pool, name, "=", value, ";", NULL);
    ap_table_set(r->headers_out, "Set-Cookie", cookie);
    return OK;
}


static table *parse_cookie(request_rec *r)
{
    const char *raw_cookie, *token, *key, *value;
    table *cookie;

    cookie = ap_make_table(r->pool, 5);
    raw_cookie = ap_table_get(r->headers_in, "Cookie");
    if (raw_cookie == NULL)
        return cookie;

    while (*raw_cookie && (token = ap_getword(r->pool, &raw_cookie, ' '))) {
         char *semi;

	 key   = ap_getword(r->pool, &token, '=');
         value = ap_getword(r->pool, &token, '=');
         if ((semi = strrchr(value, ';')) != NULL)
             *semi = '\0';

         ap_table_set(cookie, key, value);
    }
    return cookie;
}


static int has_valid_cookie(request_rec *r)
{
    deny_deeplink_rec *conf;
    table *cookie_jar;
    const char *name;
    const char *cookie;
    char *timestr;
    const char *embed_mac, *compute_mac;
    time_t timeout;

    conf = ap_get_module_config(r->per_dir_config, &deny_deeplink_module);
    name = conf->cookie_name ? conf->cookie_name : DEFAULT_COOKIE_NAME;
    cookie_jar = parse_cookie(r);
    cookie = ap_table_get(cookie_jar, name);
    if (cookie == NULL) {
        return 0; 
    }

    /* check HMAC */
    if (strlen(cookie) <= SHA_DIGESTSIZE * 2) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                      "shortest HMAC message: %u byte",
                      (unsigned int)strlen(cookie));
        return 0;
    }
    embed_mac = ap_pstrndup(r->pool, cookie, SHA_DIGESTSIZE * 2);
    timestr   = ap_pstrndup(r->pool, cookie + SHA_DIGESTSIZE * 2,
                            strlen(cookie) - SHA_DIGESTSIZE * 2);
    compute_mac = get_hmac_string(r->pool, timestr,
                                  conf->secret ? conf->secret : DEFAULT_SECRET_KEY);
    if (strcmp(embed_mac, compute_mac) != 0) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "disagreement HMAC");
        return 0;
    }

    /* check timeout */
    if (timestr == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "null timestamp");
        return 0;	    
    }
    timeout = atoi(timestr);
    /* is timeout */
    if (time(NULL) >= timeout) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "deny deeplink: timeout");
        return 0;
    }

    return 1;
}

static int check_deeplink(request_rec *r)
{
    deny_deeplink_rec *conf;

    if (r->main != NULL)
        return DECLINED;

    conf = ap_get_module_config(r->per_dir_config, &deny_deeplink_module);
    if (conf->starting_point == UNDEFINED) {
        return DECLINED;
    }

    if (conf->starting_point == ENABLE) {
        (void)set_cookie(r);
        /* disable conditional response */
	ap_table_unset(r->headers_in, "If-Modified-Since");
        ap_table_unset(r->headers_in, "If-Unmodified-Since");
        ap_table_unset(r->headers_in, "If-Match");
	ap_table_unset(r->headers_in, "If-None-Match");
	return OK;
    }
    if (conf->enable == DISABLE) {
        return DECLINED;
    }

    if (has_valid_cookie(r)) {
        /* is not deep link */
        return OK;
    }

    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "deny deeplink: redirect to '%s'",
                  conf->redirect_url ? conf->redirect_url : DEFAULT_REDIRECT_URL);
    ap_table_set(r->headers_out, "Location",
                 conf->redirect_url ? conf->redirect_url : DEFAULT_REDIRECT_URL);
    return REDIRECT;
}


/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT deny_deeplink_module = {
    STANDARD_MODULE_STUFF, 
    init_module,           /* module initializer                  */
    create_dir_config,     /* create per-dir    config structures */
    merge_dir_config,      /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    commands,              /* table of config file commands       */
    NULL,                  /* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    check_deeplink,        /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    NULL,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
#ifdef EAPI
   ,NULL,                  /* EAPI: add_module                    */
    NULL,                  /* EAPI: remove_module                 */
    NULL,                  /* EAPI: rewrite_command               */
    NULL                   /* EAPI: new_connection                */
#endif
};

